(* 
 * Copyright (c) 2001, 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: gonttypes.ml,v 1.25 2002/06/04 11:54:02 malekith Exp $
 *
 * Module implementation: Gonttypes
 * Description: The type checker. Term unification is used to check whatever
 * given two types are compatible. If we use any kind of subtyping this
 * will gonna have to be directed somewhat...
 *)

open Gonttree
open Gontutil
open Gontbind

(* Pattern ranges for checking pattern matching exhaustive *)
type patrange = 
    AlphaRange                                    (* Match all patterns *)
  | TupleRange of patrange list list
  | UnionRange of (symbol list) * (patrange list)
  | IntRange
  | StringRange
  | FloatRange
  | BoolRange of bool
;;

exception Cannot_join

exception Type_error of string

let sorry s = raise (Type_error s)
;;

let rec string_of_type_list tt =
  String.concat ", " (List.map string_of_type tt)
and string_of_fr s =
  (string_of_type s.fr_type) ^ " " ^
  (match s with
   | {fr_sym = Some sym} -> sym.sym_name
   | {fr_name = n} -> n) ^ "; "
and string_of_type_name_list l =
  (String.concat "" (List.map string_of_fr l))
and string_of_fv l =
  match l with
      [] -> ""
    | _ -> "<" ^ (String.concat ", " (List.map (fun s -> "'" ^ s) l)) ^ ">" 
and string_of_type t =
  match t with
      Void -> "void"
    | Int -> "int"
    | Float -> "float"
    | Bool -> "bool"
    | String -> "string"
    | Exn -> "exn"
    | Func f -> 
        (string_of_fv f.ft_tyvars) ^ 
        "*(" ^ (string_of_type f.ft_ret_type) ^ " ("
        ^ (string_of_type_list f.ft_parms) ^ "))" 
    | TyVar x -> ("'" ^ x)
    | Tuple l ->
        "*[" ^ (string_of_type_list l) ^ "]"
    | Struct l ->
        "struct { " ^ (string_of_type_name_list l) ^ "}"
    | OptStruct l ->
        "opt_struct { " ^ (string_of_type_name_list l) ^ "}"
    | Union l ->
        "union { " ^ (string_of_type_name_list l) ^ "}"
    | AbstractType -> "abstract_type"
    | NamedType n ->
        (match n.nt_parms with
            [] -> ""
          | l -> "<" ^ (String.concat ", " (List.map string_of_type l)) 
                     ^ ">") ^
        (if n.nt_module = "" then
           n.nt_name
         else
           n.nt_module ^ "::" ^ n.nt_name) 
;;

let rec string_of_symbol sym =
  sym.sym_name
and string_of_stmt ind s =
  let make_ind ind =  String.make (ind * 2) ' ' in
  let list_map f l = String.concat "" (List.map (f (ind + 1)) l) in
  let str = 
    let string_of_cc ind {cc_pattern = p; cc_stmt = s} =
      (make_ind ind) ^ "case " ^ (string_of_pattern p) ^ ":\n" ^
      (string_of_stmt (ind + 1) s) in
    match s.raw_stmt with
    | Skip -> 
        "skip"
    | Exp e -> 
        string_of_exp e
    | Return e -> 
        "return " ^ (string_of_exp e)
    | Raise e -> 
        "raise " ^ (string_of_exp e)
    | ReturnVoid -> 
        "return"
    | VarDef {var_symbol = Some sym} ->
        "/*VarDef*/ " ^ (string_of_symbol sym)
    | VarDef {var_name = n; var_type = t} ->
        "/*NoSymVarDef*/ " ^ n ^ ":" ^ (string_of_type t)
    | Block ss ->
        "{\n" ^ list_map string_of_stmt ss ^
        (make_ind ind) ^ "}"
    | Pair (s1, s2) ->
        "{ /* pair */\n" ^ list_map string_of_stmt [s1;s2] ^
        (make_ind ind) ^ "} /* pair */"
    | Switch (e, cc) ->
        "switch " ^ (string_of_exp e) ^ " {\n" ^
        list_map string_of_cc cc ^ (make_ind ind) ^ "}"
    | Try t -> 
        "try {\n" ^ (string_of_stmt (ind + 1) t.try_body) ^
        (make_ind ind) ^ "} with {\n" ^ 
        list_map string_of_cc t.try_with ^ 
        (make_ind ind) ^ "} finally {\n" ^
        (string_of_stmt (ind + 1) t.try_finally) ^
        (make_ind ind) ^ "}"
    | Let (pel, s) ->
        "let\n" ^ list_map 
          (fun ind (p, e) -> 
              (make_ind ind) ^ (string_of_pattern p) ^ " = " ^ 
              (string_of_exp e) ^ "\n") pel ^
        (make_ind ind) ^ "in {\n" ^ (string_of_stmt (ind+1) s) ^
        (make_ind ind) ^ "} /* let */\n"
    | If (e, s1, s2) ->
        "if " ^ string_of_exp e ^ "{ /* if */\n" ^ 
        (string_of_stmt (ind + 1) s1) ^
        (make_ind ind) ^ "} else {\n" ^ (string_of_stmt (ind + 1) s2) ^
        (make_ind ind) ^ "} /* if */"
    | Loop {ll_body = {raw_stmt = Pair ({raw_stmt = BreakIfNot c}, b)};
            ll_inc = {raw_stmt = Skip}} ->
        "while " ^ (string_of_exp c) ^ " {\n" ^
        (string_of_stmt (ind+1) b) ^ 
        (make_ind ind) ^ "}"
    | Loop {ll_body = {raw_stmt = Pair (b, {raw_stmt = BreakIfNot c})};
            ll_inc = {raw_stmt = Skip}} ->
        "do {\n" ^
        (string_of_stmt (ind+1) b) ^ 
        (make_ind ind) ^ "} while " ^ (string_of_exp c)
    | Loop b ->
        "loop {\n" ^
        (string_of_stmt (ind+1) b.ll_body) ^ 
        (make_ind ind) ^ "} uppon_continue {\n" ^
        (string_of_stmt (ind+1) b.ll_inc) ^ "}"
    | Break {br_continue = false; br_sym = Some sym} ->
        "break " ^ (string_of_symbol sym)
    | Break {br_continue = true; br_sym = Some sym} ->
        "continue " ^ (string_of_symbol sym)
    | Break {br_continue = false; br_label = n} ->
        "/*unbound*/ break " ^ n
    | Break {br_continue = true; br_label = n} ->
        "/*unbound*/ continue " ^ n
    | BreakIfNot e ->
        "if (!" ^ (string_of_exp e) ^ ") break"
    | Open m -> "open " ^ m 
  in
    (make_ind ind) ^ str ^ "\n"
and string_of_exp e =
  match e.raw_exp with
  | SimpleBinOp (o, e1, e2) ->
    let mid =
      match o with
      | Plus -> "+"
      | Minus -> "-"
      | Mult -> "*"
      | Div -> "/"
      | Mod -> "%"
      | Assign -> "="
      | Eq -> "=="
      | Ne -> "!="
      | Lt -> "<"
      | Gt -> ">"
      | Ge -> ">="
      | Le -> "<=" 
      | AndIf -> "&&"
      | OrIf -> "||"
      in
    "(" ^ (string_of_exp e1) ^ " " ^ mid ^ " " ^ (string_of_exp e2) ^ ")"
  | SimpleUnOp (o, e) ->
    let op =
      match o with
      | Negate -> "-"
      | PostDec -> "(post)--"
      | PostInc -> "(post)++"
      | PreDec -> "--"
      | PreInc -> "++"
    in
      "(" ^ op ^ " " ^ (string_of_exp e) ^ ")"
  | Ref {ref_symbol = Some sym} -> "/*ref*/ " ^ (string_of_symbol sym)
  | Ref {ref_name = s} -> s
  | CrossModuleRef {xmr_symbol = Some sym} ->
      "/*xmr*/ " ^ (string_of_symbol sym)
  | CrossModuleRef {xmr_module = m; xmr_name = n} ->
      "/*unbound xmr*/ " ^ (m ^ "::" ^ n)
  | Call (e, p) ->
      (string_of_exp e) ^ "(" ^ 
      (String.concat ", " (List.map string_of_exp p)) ^ ")"
  | IntConst s -> s
  | BoolConst(true) -> "true"
  | BoolConst(false) -> "false"
  | FloatConst s -> s
  | VoidConst -> "()"
  | NullConst -> "null"
  | StringConst s -> "\"" ^ (String.escaped s) ^ "\""
  | TupleExp el ->
      "[" ^ (String.concat ", " (List.map string_of_exp el)) ^ "]"
  | UnionExp {ur_sym = Some sym; ur_exp = e} ->
      "/*union*/ " ^ (string_of_symbol sym) ^ "[" ^
        (string_of_exp e) ^ "]"
  | UnionExp {ur_module = m; ur_name = n; ur_exp = e} ->
      "/*unbound union*/ " ^ m ^ "::" ^ n ^ "[" ^
        (string_of_exp e) ^ "]"
  | Lambda {lambda_ret_type = r;
            lambda_parms = p;
            lambda_body = b;
            lambda_tyvars = tyv} ->
      "fun <" ^ (String.concat ", " (List.map (fun t -> "'" ^ t) tyv)) ^ 
          "> " ^ (string_of_type r) ^ " (" ^ 
        (String.concat ", " (List.map 
                      (fun p -> 
                         match p with 
                         | {parm_symbol = Some sym} -> string_of_symbol sym
                         | {parm_name = n} -> "/*unbound*/ " ^ n) p)) ^ 
       ") {\n" ^ 
       (string_of_stmt 1 b) ^ "}"
  | StructExp sis -> 
     let string_of_si si =
       (match si with
        | {si_sym = Some s} -> string_of_symbol s
        | {si_module = m; si_name = n} -> m ^ "::" ^ n) 
       ^ " = " ^
       (string_of_exp si.si_expr) in
     "({ " ^ (String.concat ", " (List.map string_of_si sis)) ^ " })"
  | DotExp {de_expr = e; de_sym = Some sym} ->
     "(" ^ (string_of_exp e) ^ ")." ^ (string_of_symbol sym)
  | DotExp {de_expr = e; de_id = n} ->
     "(" ^ (string_of_exp e) ^ ")." ^ n
and string_of_pattern p =
  match p.raw_pattern with
  | VoidPattern -> "_"
  | VarPattern {vp_symbol = Some sym} -> "/*vp*/ " ^ (string_of_symbol sym)
  | VarPattern {vp_name = n} -> "/*unbound vp*/ " ^ n
  | ExprPattern e -> string_of_exp e
  | TuplePattern {tpr_patterns = p} -> 
      "[" ^ (String.concat ", " (List.map string_of_pattern p)) ^ "]"
  | UnionPattern {upr_sym = Some sym; upr_pattern = pat} ->
      "/*up*/ " ^ (string_of_symbol sym) ^ "[" ^ 
      (string_of_pattern pat) ^ "]"
  | UnionPattern {upr_module = m; upr_name = n; upr_pattern = pat} ->
      "/*unbound up*/ " ^ (m ^ "::" ^ n) ^ "[" ^ 
      (string_of_pattern pat) ^ "]"
;;

(* More entry points for debugger. *)
let print_type f t =
  Format.pp_print_string f (string_of_type t)
let print_exp f e =
  Format.pp_print_string f (string_of_exp e)
let print_pat f p =
  Format.pp_print_string f (string_of_pattern p)
let print_stmt f s =
  Format.pp_print_string f (string_of_stmt 0 s)
let print_symbol f s =
  Format.pp_print_string f (string_of_symbol s)
;;

(* o[x/t] *)
let rec subst (x, tt) o =
  let f t =
    match t with
          TyVar y when x = y -> tt
        | TyVar _ | NamedType _ | Int | Void | Bool | Float | String | Exn
        | AbstractType | Tuple _ | Func _ | Struct _ | OptStruct _ 
        | Union _ -> t in
  type_walk f o
;;

exception Is_FV

let is_fv v t =
  let f t =
     match t with
        TyVar x when v = x -> raise Is_FV
      | TyVar _ | Int | Void | Bool | Float | String | Exn
      | Func _ | Tuple _ | Union _ | Struct _ | OptStruct _ 
      | NamedType _ | AbstractType -> t in
  try
    ignore(type_walk f t);
    false
  with Is_FV -> true
;;

(* Unfortunately this function cannot be recursive, because types can
 * by cyclic... *)
let rec instance_value tt =
  let r = 
    match tt with
        NamedType ({nt_tydef = Some tydef; 
                    nt_parms = p} as n) ->
          (try
            let mapping = List.map2 (fun x y -> (x, y)) tydef.td_tyvars p in
            List.fold_left (fun o (x, t) -> subst (x, t) o) 
                           tydef.td_type 
                           mapping
          with Invalid_argument _ ->
            err (n.nt_loc, xf "type has %d parameters, but it was instanced \
                               with %d arguments" 
                              (List.length tydef.td_tyvars) 
                              (List.length p)))
      | NamedType {nt_tydef = None} -> assert false 
      | t -> t in
   match r with
   | NamedType {nt_tydef = Some _} -> instance_value r
   | t -> t
;;

(* Term unification, the standard, given-from-God, algorithm :^) *)
let rec mgu (o : (id * ty) list) rrs =
  match rrs with
      r::rs ->
          (match r with
              (TyVar(x), TyVar(y)) when x = y -> 
                  mgu o rs
            | (TyVar(x), t) 
            | (t, TyVar(x)) when not (is_fv x t) ->
                  mgu ((x, t) :: (List.map (fun (xx, tt) -> 
                                                (xx, subst (x, t) tt)) o))
                                         (* Horrible error was here... *)
                      (List.map (fun (t1, t2) -> (subst (x, t) t1, 
                                                  subst (x, t) t2)) rs)
            | (Int, Int) | (Bool, Bool) | (Void, Void)
            | (Float, Float) | (String, String) | (Exn, Exn) -> 
                  mgu o rs
            | ((Func {ft_ret_type=r1; ft_parms=p1} as x), 
               (Func {ft_ret_type=r2; ft_parms=p2} as y)) ->
                  if List.length p1 != List.length p2 then
                    sorry (xf "cannot unify function types %s and %s \
                               with different number of parameters"
                              (string_of_type x)
                              (string_of_type y))
                  else
                    mgu o ((r1, r2) :: 
                           (List.map2 (fun a b -> (a, b)) p1 p2) @ rs)
            | ((Tuple l1 as x), (Tuple l2 as y)) ->
                if List.length l1 != List.length l2 then
                  sorry (xf "cannot unify tuple types %s and %s of different \
                             arity" 
                            (string_of_type x)
                            (string_of_type y))
                else
                  mgu o ((List.map2 (fun a b -> (a, b)) l1 l2) @ rs)
            | (OptStruct [], OptStruct _)
            | (OptStruct _, OptStruct []) ->
                (* OptStruct [] is type of null pointer, so it unify with
                 * any OptStruct. *)
                mgu o rs
            | ((Struct(l1) as x), (Struct(l2) as y)) 
            | ((Union(l1) as x), (Union(l2) as y)) 
            | ((OptStruct(l1) as x), (OptStruct(l2) as y)) ->
                let cmp_fr f1 f2 =
                  match (f1, f2) with
                  | ({fr_sym = Some s1},
                     {fr_sym = Some s2}) when (sym_eq s1 s2) -> ()
                  | _ -> raise (Invalid_argument "cmp_fr") in
                (try
                  List.iter2 cmp_fr l1 l2
                 with Invalid_argument _ -> 
                        sorry (xf "cannot unify %s with %s, since they \
                                   have different struct/union fields"
                                   (string_of_type x)
                                   (string_of_type y)));
                  mgu o ((List.map2 (fun {fr_type = t1} {fr_type = t2} -> 
                                                    (t1, t2)) l1 l2) @ rs)
            (* First of all, try to unify at names level. *)
            | (NamedType {nt_tydef = t1; nt_parms = p1},
               NamedType {nt_tydef = t2; nt_parms = p2}) 
                 when (match t1, t2 with
                       | Some x, Some y ->
                         x.td_module = y.td_module && x.td_name = y.td_name
                       | _ -> false) ->
                 mgu o ((List.map2 (fun x y -> (x, y)) p1 p2) @ rs)
            (* But if it fails, try fallback to actual types. *)
            | (((NamedType _) as x), y) 
            | (x, ((NamedType _) as y)) ->
                  mgu o ((instance_value x, instance_value y) :: rs)
            | (x, y) ->
                    sorry (xf "cannot unify %s with %s" 
                              (string_of_type x)
                              (string_of_type y)))
                                     
    | [] -> o
;;

type ty_context = {ctx_ret_type: ty;   (* Return type of function. *)
                   ctx_tyvars: id list (* Variables functions is parametrized
                                        * over. *)
                  }

(* Oh well... this function is misleading. *)
let fresh_vars t =
  let f t =
    match t with
          Func {ft_tyvars = ((_::_) as tv)} -> 
            List.fold_left (fun t x -> subst (x, 
                                              TyVar(seq_name "$")) t) 
                           t tv
        | Func _ | TyVar _ | NamedType _ | Int | Void | Bool | Float | String 
        | AbstractType | Tuple _ | Struct _ | OptStruct _  | Exn
        | Union _ -> t in
  type_walk f t
  
(* Check whatever s and t are compatible in given context c.
 * Return MGU. *)
let checker c s t =
  let check_if_nok lst (v, t) =
    let swaper (u, v) n =
      if n = u then v else if n = v then u else n in
    let term_swaper (u, v) t =
      match t with TyVar x -> TyVar (swaper (u, v) x) | t -> t in
    let swap (u, v) (n, t)  =
      (swaper (u, v) n, type_walk (term_swaper (u, v)) t) in
    if List.mem v c.ctx_tyvars then
      match t with
          TyVar x when not (List.mem x c.ctx_tyvars) -> 
            List.map (swap (x, v)) lst
        | _ -> sorry (xf "'%s has to be %s, while it should be \
                          assumption-free" v (string_of_type t))
    else lst in
  let m = mgu [] [(t, s)] in
  List.fold_left check_if_nok m m

(* Check if t < s. *)
let subtype c s t =
  (* I guess this is wrong (fresh_vars doesn't do exactly what it's
   * expected to...) *)
  ignore(checker c (fresh_vars t) (fresh_vars s))
  
let global_subtype s t =
  subtype {ctx_ret_type = AbstractType; ctx_tyvars = []} s t
  
let typecheck_call c f at =
  let ft = 
    match f with
        Func {ft_parms = x} -> x
      | _ -> sorry (xf "called object is not a function \
                        (it has non-functional type %s)"
                               (string_of_type f)) in
  if (List.length ft) != (List.length at) then
    sorry (xf "there is %d formal parameters and %d actual parameters in call"
              (List.length ft) (List.length at));
  let var = seq_name "$" in
  let a' = Func {ft_ret_type = TyVar(var); ft_parms = at; ft_tyvars=[var]} in
  let m = checker c a' (fresh_vars f) in
  try
    let (_, t) = 
      (try
        List.find (fun (x, _) -> x = var) m 
       with Not_found ->
         let (n, _) = List.find (fun (_, x) -> 
                                        match x with 
                                            TyVar(y) when y = var -> true
                                          | _ -> false) m in 
         (n, TyVar n)) in
    match t with
         TyVar k -> 
           if k.[0] == '$' then
             t
           (* this can happen for void -> 'a like functions
             sorry (xf "what the heck? return value of call is '%s" k)
             *)
           else
             t
       | _ -> t
  with Not_found ->
    (* hmm... shouldn't we return TyVar(var) here? *)
    sorry "return type of call is unspecified"
;;

let rec lvalue_p e =
  match e.raw_exp with
      (* Of course this might be wrong... We need some 'const' notation. *)
      Ref r -> ()
    | DotExp t -> ()
    | TupleExp t -> 
       let _ = List.map lvalue_p t  in
       ()
    | _ -> sorry "expression is not a proper lvalue"
;;

let rec strip_off ({ctx_tyvars = t} as g) l =
  match l with
      x::xs -> 
        if List.mem x t then
          strip_off g xs
        else
          x :: (strip_off g xs)
    | [] -> []
    
let stripper g t =
  let f t =
    match t with
          Func ({ft_tyvars = tt} as f) -> 
            f.ft_tyvars <- strip_off g tt;
            Func f
        | TyVar _ | NamedType _ | Int | Void | Bool | Float | String 
        | AbstractType | Tuple _  | Struct _ | OptStruct _ | Exn
        | Union _ -> t in
  ignore(type_walk f t)

(* Look for named field in union type, and return type of this particular
 * field. (We cannot use sym.sym_type here, as it can be too general. *)
let type_of_union_field_in tex sym =
  let {fr_type = ft} = 
    List.find (fun f -> 
                 match f with
                 | {fr_sym = Some s} when (sym_eq s sym) -> true
                 | _ -> false)
              (match tex with Union l -> l | _ -> assert false) in
  ft

let rec typecheck_exp (g, e) =
  let subtype_p g t1 t2 =
    try
      subtype g t1 t2; true
    with Type_error _ -> false in
  try
    let (g', t) =
      match e.raw_exp with
          SimpleBinOp(op, e1, e2) ->
            let (g1, t1) = typecheck_exp (g, e1) in
            let (g2, t2) = typecheck_exp (g1, e2) in
            (match op with
              | Mod -> 
                  subtype g2 Int t1;
                  subtype g2 Int t2;
                  (g2, Int)
              | Plus | Minus | Div | Mult ->
                  if (subtype_p g2 Int t1) then
                    begin
                      subtype g2 Int t1;
                      subtype g2 Int t2;
                      (g2, Int)
                    end
                  else if (subtype_p g2 Float t1) then
                    begin
                      subtype g2 Float t1;
                      subtype g2 Float t2;
                      (g2, Float)
                    end
                  else
                    begin
                      match op with
                      | Plus when (subtype_p g2 String t1) ->
                          subtype g2 String t1;
                          subtype g2 String t2;
                          (g2, String)
                      | _ -> sorry (xf "-, / and * are supported only \
                                        on bools and floats, + also on \
                                        strings, not on (%s, %s)"
                                        (string_of_type t1)
                                        (string_of_type t2))
                    end
              | Assign ->
                  subtype g2 t1 t2;
                  lvalue_p e1;
                  (g2, t1)
              | Eq | Ne ->
                  (if subtype_p g2 Bool t1 then
                    subtype g2 Bool t2
                  else if subtype_p g2 Int t1 then
                    subtype g2 Int t2
                  else if subtype_p g2 Float t1 then
                    subtype g2 Float t2
                  else if subtype_p g2 String t1 then
                    subtype g2 String t2
                  else 
                    match ((instance_value t1), (instance_value t2)) with
                    | (OptStruct [], OptStruct x)
                    | (OptStruct x, OptStruct []) -> ()
                    | _ ->
                      sorry (xf "`!=' and `==' operators supported only on \
                                 bool, int, float, string types and null \
                                 pointers, not on (%s, %s)"
                              (string_of_type t1)
                              (string_of_type t2)));
                  (g2, Bool)
              | Lt | Gt | Le | Ge ->
                  if subtype_p g2 Int t1 then
                    subtype g2 Int t2
                  else if subtype_p g2 Float t1 then
                    subtype g2 Float t2
                  else if subtype_p g2 String t1 then
                    subtype g2 String t2
                  else
                    sorry (xf "`<', '>', '<=' and `>=' operators supported \
                               only on int, string and float types, not \
                               on (%s, %s)"
                              (string_of_type t1)
                              (string_of_type t2));
                  (g2, Bool)
               | AndIf | OrIf ->
                  subtype g2 Bool t1;
                  subtype g2 Bool t2;
                  (g2, Bool)
              )
        | SimpleUnOp(op, e) ->
            let (g', t) = typecheck_exp (g, e) in
            (match op with
               | Negate -> 
                 if subtype_p g' Int t then
                   (g', Int)
                 else if subtype_p g' Float t then
                   (g', Float)
                 else
                   sorry (xf "negation operator only supported on ints and \
                              floats, not on %s" (string_of_type t))
               | PreDec | PreInc | PostDec | PostInc ->
                 subtype g' Int t;
                 lvalue_p e;
                 (g', Int))
        (* This is a bit special, convert Nil, List::Nil to Nil[], 
         * List::Nil[] (where Nil is part of some union, of course). *)
        | Ref {ref_symbol = Some sym}
        | CrossModuleRef {xmr_symbol = Some sym} 
                when (has_property sym IsUnionField || 
                      has_property sym IsException) ->
             (e.raw_exp <- 
                 UnionExp {ur_exp = {raw_exp = VoidConst;
                                     exp_loc = e.exp_loc;
                                     exp_type = Some Void};
                           ur_module = "";
                           ur_name = "";       (* We don't need name here, *)
                           ur_sym = Some sym}; (* as we have the symbol. *)
              typecheck_exp (g, e))
        | Ref {ref_symbol = Some sym} ->
           (let t = sym.sym_type in
            (* Drop type variables from locals. This is because in 
             * 'a f( *('a)->'a x) { ... } f has type 'a->'a, where
             * 'a is some given type (it is not free). *)
            if sym.sym_fnc_level > 0 then stripper g t;
            (g, t))
        | Ref {ref_symbol = None} -> assert false
        | Call(f, act) ->
            let (g', t) = typecheck_exp (g, f) in
            let (g'', tact) = list_apply typecheck_exp (g', act) in
            let ret_type = typecheck_call g'' t tact in
            (* DEBUG: *)
            (* warn(e.exp_loc, "rettype = " ^ (string_of_type ret_type)); *)
            (g'', ret_type)
        | IntConst _ -> (g, Int)
        | BoolConst _ -> (g, Bool)
        | FloatConst _ -> (g, Float)
        | StringConst _ -> (g, String)
        | VoidConst -> (g, Void)
        | NullConst -> (g, OptStruct [])
        | CrossModuleRef {xmr_symbol = Some sym} -> (g, sym.sym_type)
        | CrossModuleRef {xmr_symbol = None} -> assert false
        | TupleExp tup ->
            let rec tuple_typer (g, tu) =
              match tu with
                x::xs -> 
                  let (g', t) = typecheck_exp (g, x) in
                  let (g'', t') = tuple_typer(g', xs) in
                  (g'', t::t')
              | [] -> (g, []) in
            let (g', t) = tuple_typer (g, tup) in
            (g', Tuple t)
        | UnionExp {ur_sym = None} -> assert false
        | UnionExp ({ur_sym = Some sym} as un) -> 
            (*
              u { 'a X; 'a Y }

              switch (X[1.0]) {
              case X[1.0]: ....
              case Y[1]: // type error
              }
            *)
            if has_property sym IsUnionField then
              let (g', t) = typecheck_exp (g, un.ur_exp) in
              let ut = fresh_vars (union_typedef sym).td_type in
              let ft = type_of_union_field_in ut sym in
              let a = mgu [] [(ft, t)] in
              let ut' = 
                List.fold_left (fun ut (i, t) -> subst (i, t) ut) ut a in
              (g', ut')
            else if has_property sym IsException then
              let (g, t) = typecheck_exp (g, un.ur_exp) in
              let _ = subtype g sym.sym_type t in
              (g, Exn)
            else 
              sorry(xf "identifier `%s' is not part of any union"
                       sym.sym_name)
        | StructExp sis ->
            let type_of_typedef td =
                NamedType {nt_module = td.td_module; 
                           nt_name = td.td_name;
                           nt_tydef = Some td;
                           nt_parms = List.map 
                                        (fun _ -> TyVar(seq_name "$si"))
                                        td.td_tyvars;
                           nt_loc = e.exp_loc} in
             let (g, possible_types) = 
               match sis with
               | {si_sym = Some {sym_typedefs = td}}::_ -> 
                    typecheck_sis g (List.map type_of_typedef td) sis
               | _ -> assert false in
             let t = 
               match possible_types with
               | [t] -> chini_si t sis; t
               | (_::_) as ts ->
                   sorry ("structure initializer can have more the one \
                           type, possible types: " ^
                          (String.concat ", " (List.map string_of_type ts)))
               | [] -> assert false (* should be handled by typecheck_sis,
                                     * where we can give more informative 
                                     * message. *)
             in 
               (g, t)
        | DotExp de ->
            let (g, t) = typecheck_exp (g, de.de_expr) in
            let lst = 
              match (instance_value t) with
              | Struct l 
              | OptStruct l -> l
              | _ ->
                  sorry ("left-hand side of `.' is of non-struct type " ^
                         (string_of_type t)) in
            let s =
              match lst with
              | {fr_sym = Some s}::_ -> s 
              | _ -> assert false in
            let qn = sym_module s  ^ "::" ^ de.de_id in
            let rec find_fr lst =
              match lst with
              | ({fr_sym = Some {sym_name = n}} as x)::_ when (n = qn) -> x
              | _::xs -> find_fr xs 
              | [] -> sorry ("type " ^ (string_of_type t) ^ 
                             " has no field named " ^ qn) in
            let fr = find_fr lst in
            let ty =
              match fr with 
              | {fr_sym = Some sym; fr_type = t} -> 
                  de.de_sym <- Some sym;
                  t
              | _ -> assert false in
            (g, ty)
        | Lambda l  ->
            (* Here we need to strip off type variables, that was
             * put here by stupid parser. *)
            List.iter (fun {parm_type=t} -> stripper g t) l.lambda_parms;
            let tyvars = (strip_off g l.lambda_tyvars) @ g.ctx_tyvars in
            ignore(typecheck_stmt ({g with 
                                    ctx_ret_type = l.lambda_ret_type;
                                    ctx_tyvars = tyvars},
                                   l.lambda_body));
            (g, Func {ft_ret_type = l.lambda_ret_type;
                      ft_parms = List.map (fun {parm_type=t} -> t)
                                          l.lambda_parms;
                      ft_tyvars = tyvars}) in
    e.exp_type <- Some t;
    (g', t)
  with Type_error s -> raise (Syntax_error (e.exp_loc, s))

(* Typecheck given structure initializer. possible_types is list of types
 * this expression might have. sis is list of struct_init records. Return
 * (context, list of possible types). *)
and typecheck_sis g possible_types sis =
  match sis with
  | {si_sym = Some sym; si_expr = exp; si_loc = loc}::sis ->
      (* Check whaterver given (sym, exp) pair could be part of given
       * named_type. Return Some named_type and context after substitutions,
       * or None and context if it's impossible. *)
      let check_one_type g named_type lastone =
        match (instance_value named_type) with
        | OptStruct lst 
        | Struct lst ->
           let (ex_g, tex) =
              match exp with
              | {exp_type = Some t} -> (g, t)
              | _ -> typecheck_exp (g, exp) in
           (try
              let sub o (n, t) =
                let repl (n, t) =
                  match o with
                  | NamedType ({nt_parms = p} as r) ->
                     let rec sub_in v =
                       match v with
                       | (TyVar x)::xs when (x = n) ->
                           t::xs
                       | x::xs -> 
                           x::(sub_in xs) 
                       | [] -> assert false in
                     NamedType {r with nt_parms = sub_in p} 
                  | _ -> assert false in
                match (n, t) with
                | (_, TyVar k)
                   when ((String.length k) > 3 &&
                         (String.sub k 0 3) = "$si") ->
                     repl (k, TyVar n)
                | (n, _)
                   when ((String.length n) > 3 &&
                         (String.sub n 0 3) = "$si") -> 
                      repl (n, t)
                | (n, t) ->
                      o
                    (*
                    sorry (xf "type variable '%s should be assumption-free \
                               while it is deemed to be %s" 
                               n 
                               (string_of_type t)) *) in
              let fr = 
                 try 
                   List.find (fun fr -> 
                                match fr with 
                                | {fr_sym = Some s} -> sym_eq sym s
                                | _ -> assert false) lst
                 with Not_found ->
                   if lastone then
                     err (loc, xf "%s has no field named %s"
                                  (string_of_type named_type)
                                  sym.sym_name)
                   else
                     raise Not_found in
              let substituted_type = 
                 try
                   List.fold_left sub 
                                  named_type 
                                  (mgu [] [(fr.fr_type, tex)])
                 with Type_error s ->
                   if lastone then
                     err (loc, xf "field %s of %s cannot have this type: %s"
                                   sym.sym_name 
                                   (string_of_type named_type) s)
                   else
                     raise (Type_error s) in
              (ex_g, Some substituted_type)
            with Type_error _ | Not_found -> 
                (ex_g, None))
        | _ -> assert false in
    let rec filter_types r g possible_types =
      match possible_types with
      | x::xs ->
          let lastone =
            match (xs, r) with
            | ([], []) -> true
            | _ -> false in
          (match (check_one_type g x lastone) with
           | (g, Some t) ->
              filter_types (t::r) g xs
           | (g, None) ->
              filter_types r g xs)
      | [] -> (g, r) in
    let (g, possible_types) = filter_types [] g possible_types in
    typecheck_sis g possible_types sis
  | {si_sym = None}::_ -> assert false
  | [] -> (g, possible_types)

(* Check whatever all fields in t are initialized by sis. *)
and chini_si t sis =
  let rec chini_si_rec flds sis =
    match sis with
    | {si_sym = Some s; si_loc = loc}::xs ->
      if List.memq s flds then
        chini_si_rec (List.filter (fun f -> f != s) flds) xs
      else
        err (loc, "field " ^ s.sym_name ^ " initialized more then once")
    | {si_sym = None}::_ -> assert false
    | [] -> flds in
  let lst = 
    match (instance_value t) with
    | Struct l
    | OptStruct l -> l
    | _ -> assert false in
  let syms = List.map (fun f -> match f with
                                | {fr_sym = Some s} -> s
                                | _ -> assert false) lst in
  let unini = chini_si_rec syms sis in
  match unini with
  | [] -> ()
  | x -> sorry ("some struct fields are not initialized: " ^
                (String.concat ", " (List.map (fun s -> s.sym_name) x)))

and typecheck_stmt (g, s) =
  let rec typecheck_patterns te (g, pl) =
    match pl with
    | {cc_pattern = p; cc_stmt = s}::xs ->
        (* instance_value is called by typecheck_pattern now. *)
        let _ = typecheck_pattern(te, g, p) in
        let (g, _) = typecheck_stmt(g, s) in
        typecheck_patterns te (g, xs)
    | [] -> g in
  try
    match s.raw_stmt with
        Skip -> 
          (g, Void)
      | Exp(e) -> 
          let (g', _) = typecheck_exp (g, e) in 
          (g', Void)
      | Return(e) -> 
        let (g', t) = typecheck_exp (g, e) in
        subtype g g.ctx_ret_type t; 
        (g', Void)
      | ReturnVoid ->
        subtype g g.ctx_ret_type Void;
        (g, Void)
      | Block(ss) ->
        let (g', _) = list_apply typecheck_stmt (g, ss) in
        (g', Void)
      | VarDef {var_type = t} ->
        (* Here we need to strip off type variables, that was
         * put here by stupid parser. *)
        stripper g t;
        (g, Void)
      | Open(_) ->
        (g, Void)
      | Pair(a, b) ->
        let (g', _) = typecheck_stmt (g, a) in
        let (g'', _) = typecheck_stmt (g', b) in
        (g'', Void)
      | BreakIfNot e ->
        let (g', t) = typecheck_exp (g, e) in
        subtype g Bool t;
        (g', Void)
      | Switch (e, pl) ->
          let (g, te) = typecheck_exp (g, e) in
          let g = typecheck_patterns te (g, pl) in
          pattern_exhaustive e.exp_loc [] pl;
          (g, Void)
      | Let (patel, st) ->
          let rec typecheck_let (g, pel) =
            match pel with
              (pat, ex) :: xs ->
                 let (g, te) = typecheck_exp (g, ex) in
                 let _ = typecheck_pattern(te, g, pat) in
                 typecheck_let (g, xs)
            | [] -> g in
          let g = typecheck_let (g, patel) in
          let (g, _) = typecheck_stmt(g, st) in
          (g, Void)
      | Raise e ->
          let (g, t) = typecheck_exp (g, e) in
          subtype g Exn t;
          (g, Void)
      | Try t ->
          let (g, _) = typecheck_stmt (g, t.try_body) in
          let g = typecheck_patterns Exn (g, t.try_with) in
          (* Don't call pattern exhaustive for try_with, it's rather
           * senseless. *)
          let (g, _) = typecheck_stmt (g, t.try_finally) in
          (g, Void)
      | If(e, s1, s2) ->
          let (g1, t) = typecheck_exp (g, e) in
          let (g2, _) = typecheck_stmt (g1, s1) in
          let (g3, _) = typecheck_stmt (g2, s2) in
          subtype g Bool t;
          (g3, Void)
      | Break _ -> 
        (* Break is always type-correct. *)
        (g, Void)
      | Loop {ll_body = s1; ll_inc = s2} ->
        let (g, _) = typecheck_stmt (g, s1) in
        let (g, _) = typecheck_stmt (g, s2) in
        (g, Void)
  with Type_error e -> raise (Syntax_error (s.stmt_loc, e))

and is_const ex =
  match ex.raw_exp with
  | SimpleBinOp(_, e1, e2) -> (is_const e1) && (is_const e2)
  | SimpleUnOp(_, e) -> is_const e
  | Lambda _  
  | Ref _ 
  | Call _ 
  | StructExp _ 
  | DotExp _
  | CrossModuleRef _ -> false
  | IntConst _
  | BoolConst _
  | FloatConst _
  | VoidConst 
  | NullConst
  | StringConst _ -> true
  | TupleExp tup ->
     not (List.exists (fun e -> not (is_const e)) tup)
  | UnionExp {ur_exp = e} -> is_const e

and typecheck_pattern (tex, g, pa) =
    match pa.raw_pattern with
    | VoidPattern ->
       ()
    | VarPattern {vp_symbol = Some sym} ->
       sym.sym_type <- tex
    | VarPattern {vp_symbol = None} -> assert false
    | ExprPattern const ->
       let (g, t) = typecheck_exp (g,const) in
       subtype g tex t;
       if is_const const then
         ()
       else
         err (const.exp_loc, "only constant expressions in pattern matching \
                              allowed")

    | TuplePattern {tpr_patterns = p} ->
        let np = List.length p in
        let loc = pa.pat_loc in
        (match (instance_value tex) with
         | Tuple t ->
             if List.length t != np then
               err(loc, xf "pattern is %d-tuple, value is %d-tuple"
                           np (List.length t))
             else
               List.iter2 (fun t p -> typecheck_pattern (t, g, p)) t p
         (* Tuple patter with empty list of patterns is special case --
          * pattern that only matches void value. So we allow it here. *)
         | Void when (np == 0) -> ()
         (* However 1-tuple isn't supported ;) (it might have occurred from
          * Foo[x], but [x] there is not a pattern anymore. *)
         | _ ->
             err(loc, xf "expected tuple as value to match"))
    | UnionPattern ({upr_sym = Some sym} as un) ->
        begin
          try 
            if has_property sym IsUnionField then
              let _ = subtype g (union_typedef sym).td_type tex in
              let t = type_of_union_field_in (instance_value tex) sym in
              typecheck_pattern (t, g, un.upr_pattern)
            else if has_property sym IsException then
              let _ = subtype g Exn tex in
              typecheck_pattern (sym.sym_type, g, un.upr_pattern)
            else 
              sorry (xf "identifier `%s' is not part of any union"
                         un.upr_name)
          with Type_error s -> err(pa.pat_loc, s)
        end
    | UnionPattern {upr_sym = None} -> assert false 

and pattern_exhaustive loc ran_l plist =
  match plist with
  | {cc_pattern = pat}::xs ->
      let pat_r = range_of_pattern pat in
      let (cmp, nran_l) = join_lists ran_l [pat_r] in
      (if (cmp <= 0) then 
        let lo = pat.pat_loc in
        warn (lo, "pattern clause is unused"));
      pattern_exhaustive loc nran_l xs
  | [] ->
      let hd = 
        (match ran_l with
        | x::_ -> x
        | [] -> assert false
        ) in
      if (hd <> AlphaRange) then
        let (_, example) = generate_example 0 ran_l in
        warn (loc, xf "this pattern matching isn't exhaustive, here is \
                       example of pattern which isn't matched: %s" example)
      else ()

and range_of_pattern pat =
  match pat.raw_pattern with
  | VoidPattern ->
      AlphaRange
  | VarPattern _ ->
      AlphaRange
  | ExprPattern const ->
      (match const.raw_exp with
      | IntConst _ -> IntRange
      | BoolConst b -> BoolRange b
      | FloatConst _ -> FloatRange
      | StringConst _ -> StringRange
      | _ -> AlphaRange
      )
  | TuplePattern {tpr_patterns = p} ->
      let lister x = [x] in
      let rl = List.map range_of_pattern p in
      let rll = List.map lister rl in
      TupleRange rll
  | UnionPattern {upr_sym = Some sym; upr_pattern = un_pat} ->
      let rp = range_of_pattern un_pat in
      UnionRange ([sym], [rp]) 
  | UnionPattern {upr_sym = None} -> assert false 

and generate_example eq rpats =
  let rpat = 
    match rpats with
    | x::xs -> x
    | [] -> AlphaRange in
  match rpat with
  | AlphaRange -> (eq, "_")
  | IntRange -> (1, "some int")
  | FloatRange -> (1, "some float")
  | StringRange -> (1, "some string")
  | BoolRange b -> 
      if (b && eq == 1 || not b && eq == 0) then (1, "true")
      else (1, "false")
  | UnionRange (syml, [AlphaRange]) ->
      let rec find_absent_name l1 l2 =
        match l2 with
        | x::xs ->
            let exist = List.exists (function y -> (x == y)) l1 in
            if exist then find_absent_name l1 xs
            else x
        | [] -> assert false in
     let ty = (union_typedef (List.hd syml)).td_type in
     let tsym_l = 
       match ty with
       | Union l -> List.map (function f -> 
                                 match f with 
                                 | {fr_sym = Some s} -> s
                                 | _ -> assert false) l
       | _ -> assert false in
     if (eq == 0) then
       let nm_name = find_absent_name syml tsym_l in
       let fu y =
         match y with
         | UnionRange ([x],_) when (x == nm_name) -> true
         | _ -> false in
       let absent_pat = List.filter fu rpats in
       if (absent_pat = []) then (1, nm_name.sym_name ^ " _")
       else
         generate_example 0 absent_pat
     else 
       (1, (string_of_symbol (List.hd syml)) ^ "_")
  | UnionRange ([sym], rplist) ->
      let (neq, str) = generate_example eq rplist in
      (neq, (string_of_symbol sym) ^ " " ^ str)
  | UnionRange (_, _) -> assert false
  | TupleRange rtl ->
      let rec stringer eql li =
        match li with
        | x::xs -> 
            let (eq', str) = generate_example eql x in
            let (eq'', strr) = stringer eq' xs in
            (eq'', str ^ ", " ^ strr)
        | [] -> (eql, "") in
      let (neq, st) = stringer eq rtl in
      (neq, "[" ^ st ^ "]")

(*
  old < new -> 1
  old = new -> 0
  old > new -> -1
  joined(r1, r2) -> 2
*)
and fold_tuple t =
  let fold_list l =
     let l = List.map fold_tuple l in
     match l with
     | [x] -> [x]
     | [] -> []
     | x::xs -> let (_, j) = join_lists [x] xs in j in
  match t with
  | TupleRange l ->
     let l = List.map fold_list l in
     if List.exists (fun x -> x <> [AlphaRange]) l then
       TupleRange l
     else
       AlphaRange
  | UnionRange (syms, pats) ->
      UnionRange (syms, fold_list pats)
  | AlphaRange | IntRange | StringRange
  | FloatRange | BoolRange _ -> t
  
and join_ranges oldr newr =
  match (oldr, newr) with
  | (AlphaRange, AlphaRange) -> (0, AlphaRange)
  | (AlphaRange, _) -> (-1, AlphaRange)
  | (_, AlphaRange) -> (1, AlphaRange)
  | (UnionRange (old_s, [AlphaRange]), UnionRange (new_s, [AlphaRange])) ->
     let rec uniq l1 l2 =
       match l2 with
       | x::xs -> 
         if List.memq x l1 then 
           uniq l1 xs 
         else
           uniq (x :: l1) xs
       | [] -> l1 in
     let joined = List.sort compare (uniq old_s new_s) in
     let ty = (union_typedef (List.hd old_s)).td_type in
     let needed_length = 
       match ty with
       | Union l -> List.length l
       | _ -> assert false in
     let joined_length = List.length joined in
     let l1_length = List.length old_s in
     let l2_length = List.length new_s in
     let res =
        if l1_length == joined_length && 
           l2_length == joined_length then 
             0
        else if l1_length == joined_length then
             -1
        else if l2_length == joined_length then
             1
        else
             2 in
     if joined_length == needed_length then
       (res, AlphaRange)
     else
       (res, UnionRange (joined, [AlphaRange]))
  | (UnionRange (old_s, [AlphaRange]), UnionRange ([s], _)) ->
      if List.memq s old_s then
        (-1, oldr)
      else
        raise Cannot_join
  | (UnionRange ([s], _), UnionRange (new_s, [AlphaRange])) ->
      if List.memq s new_s then
        (1, newr)
      else
        raise Cannot_join
  | (UnionRange ([old_s], old_l), UnionRange ([new_s], new_l)) -> 
      if old_s == new_s then
        let (res, l) = join_lists old_l new_l in
        (res, UnionRange ([new_s], l))
      else
        raise Cannot_join
  | (UnionRange _, UnionRange _) -> assert false
  | (TupleRange old_l, TupleRange new_l) ->
      let rec join state ret l =
        match l with
        | (o, n)::xs ->
          let (cmp, p) = join_lists o n in
          (match (cmp, state) with
           | (0, _) -> 
               join state (p :: ret) xs
           | (1, 0)
           | (1, 1)
           | (-1, 0) 
           | (-1, -1) -> 
               join cmp (p :: ret) xs
           | (2, 0) ->
               join cmp (p :: ret) xs
           | _ -> raise Cannot_join)
        | [] -> (state, fold_tuple (TupleRange (List.rev ret))) in
      join 0 [] (List.combine old_l new_l)
  | (BoolRange b1, BoolRange b2) ->
    if (b1 == b2) then (0, BoolRange b1)
    else (2, AlphaRange)
  | (IntRange, _) -> raise Cannot_join
  | (FloatRange, _) -> raise Cannot_join
  | (StringRange, _) -> raise Cannot_join
  | (_, _) -> assert false
    
and join_lists old_l new_l =
  let res = List.fold_left append_to_list old_l new_l in
  let res = List.sort compare res in
  if (res = old_l) && (res = new_l) then
    (0, res)
  else if (res = old_l) then
    (-1, res)
  else if (res = new_l) then
    (1, res)
  else
    (2, res)
  
and append_to_list old_l new_r =
  let rec append_rec res old new_r =
    match old with
    | x::xs ->
        (try
           let (_, p) = join_ranges x new_r in
           append_to_list (res @ xs) p
        with Cannot_join ->
          append_rec (x :: res) xs new_r)
    | [] -> new_r :: res in
  append_rec [] old_l (fold_tuple new_r)
;;

(* Introduce new, artificial, type variables in type definition. This helps
 * up catching errors. ('$v* should not appear in type of any expression). *)
let td_fresh_variables td =
  match td.td_type with
  | AbstractType ->
    td.td_tyvars <- List.map (fun _ -> seq_name "$v") td.td_tyvars
  | _ ->
    let map : (string, string) Hashtbl.t = Hashtbl.create 16 in
    let f t =
      match t with
      | TyVar a ->
        if not (Hashtbl.mem map a) then
          Hashtbl.add map a (seq_name "$v"); 
        TyVar (Hashtbl.find map a)
      | t -> t in
    td.td_type <- type_walk f td.td_type;
    td.td_tyvars <- 
      List.map (fun n -> 
                  try 
                    Hashtbl.find map n 
                  with Not_found ->
                    err(td.td_loc, xf "unused type variable '%s" n)) 
               td.td_tyvars
;;                                                    

(* chini_* (check if initialized) functions check whatever all variables
 * are properly initialized before they are used. They are somewhat 
 * simplistic (or conservative). This could be improved I guess. 
 *
 * `return' stmt is treated as initialization of *all* variables.
 * We maintain list of initialized variables, and walk with it down the
 * expression/stmt terms. return stmt introduces Everything node to this
 * list, it should be present in functions returning non-void, after 
 * precessing top-level stmt (each path ends with return then).
 *
 * Precessing `break' stmts in loops are somewhat tricky. They push 
 * special Breaker node with level of loop nesting to the initialized
 * variables list. Then, variables initialized in loop, are these
 * before first Breaker pushed (they are after last breaker in the list),
 * with Breaker prepended if it would break upper level loop (we don't
 * use it right now, as there is no break stmt in Gont right now..., 
 * however for and while loops are translated to loop(if (c) break; ...))
 *
 * Handling of Lambda expression is currently somewhat too conservative.
 * For example:
 *
 *      int f() 
 *      { 
 *              int m;
 *              void g() 
 *              { 
 *                      return m;       // gontc complains here
 *              }
 *              m = 5;
 *              return g();
 *      }
 *
 * I don't know how could we fix it, though...
 *)

type ini =
    Symbol of symbol
  | Breaker of int
  | Everything

type chini_ctx = {cc_initialized: ini list;
                  cc_this_stmt_initialized: ini list;
                  cc_loop_level: int;
                  cc_try_level: int}

(* Check initialization in sequence of stmts/exprs (depending on chini_one
 * functional parameter). Sequence is given in lst. *)
let chini_list chini_one gctx lst =
  let rec chini_list_tail ini ctx lst =
    match lst with
    | x::xs ->
      let ctx = chini_one ctx x in
      let ini = ctx.cc_this_stmt_initialized @ ini in
      let ctx = {ctx with cc_this_stmt_initialized = [];
                          cc_initialized = ini @ gctx.cc_initialized} in
      chini_list_tail ini ctx xs
    | [] -> ini in
  let ini = chini_list_tail [] gctx lst in
  {gctx with cc_this_stmt_initialized = ini}

let get_loop_level sym =
  assert (sym.sym_kind = LoopIdKind);
  let rec f lst =
    match lst with
    | (LoopLevel i)::_ -> i
    | _::xs -> f xs
    | [] -> assert false in
  f sym.sym_properties
  
(* Check initialization in one stmt. *)
let rec chini_stmt ctx s =
    (* Find minimum of v1 v2, treat -1 as +Infty *)
    let strange_min v1 v2 =
      if v1 == -1 then v2
      else 
        if v2 == -1 then v1
        else
          if v1 < v2 then v1
          else v2 in
    
    (* a b B1 c d B2 d e -> a b c d B1 d e *)
    (* 1: { d = 1;
            2: { e = 1;
                 break 2;
                 d = 1;
                 c = 1;
                 break 1;
                 b = 1;
               }
            a = 1;
          }
     *)     
    let normalize lst =
      let rec normalize_tail min r1 r2 lst =
        match lst with
        | (Breaker k) :: xs -> 
          let min = strange_min min k in
          normalize_tail min (r2 @ r1) [] xs
        | x :: xs ->
          normalize_tail min r1 (x :: r2) xs
        | [] ->
          match r1 with
          | [] ->
             ([], min, r2)
          | _ ->
             (r1, min, r2) in
      normalize_tail (-1) [] [] lst in
    let sort = List.sort compare in
    (* Find intersection of two given, sorted lists. *)
    let intersection lst1 lst2 =
      let rec find_is res lst1 lst2 =
        match lst1 with
        | x::xs ->
          (match lst2 with
           | y::ys ->
             let c = compare x y in
             if c < 0 then
               find_is res xs lst2
             else
               if c > 0 then
                 find_is res lst1 ys
               else
                 find_is (x::res) xs ys
           | [] -> res)
        | [] -> res in
      if List.mem Everything lst1 then 
         lst2
      else 
         if List.mem Everything lst2 then
           lst1
         else
           List.rev (find_is [] lst1 lst2) in
    (* Check initialization in mutually exclusive branches given in l. *)
    let chini_branches_no_ctx gctx l =
      let rec chini_branches_tail (abr, br, bbr) l =
        match l with
        | x::xs ->
          let ctx = chini_stmt gctx x in
          let (abr', br', bbr') = 
                normalize ctx.cc_this_stmt_initialized in
          let abr = abr' @ abr in
          let bbr = intersection bbr (sort bbr') in
          let br = strange_min br br' in
           chini_branches_tail (abr, br, bbr) xs
        | [] -> (abr, br, bbr) in
       chini_branches_tail ([Everything], -1, [Everything]) l in
    (* if foo
         i = 1;
         if bar
           break xxx;
         j = 2;
       else
         i = 1;
         j = 2;
         k = 3;
       -> ([j], xxx, [i]) *)
    let chini_branches c l = 
      let (abr, min, bbr) = chini_branches_no_ctx c l in
      let ini =
        if min == -1 then bbr
        else abr @ ((Breaker min) :: bbr) in
      {c with cc_this_stmt_initialized = ini} in
    
    match s.raw_stmt with
    | Skip 
    | VarDef _ 
    | Open _ -> ctx
    | Exp e -> chini_exp ctx e
    | Return e -> 
        let ctx = chini_exp ctx e in
        let ini = Everything :: ctx.cc_this_stmt_initialized in
        {ctx with cc_this_stmt_initialized = ini}
    | ReturnVoid ->
        {ctx with cc_this_stmt_initialized = [Everything]}
    | Block ss -> 
        chini_list chini_stmt ctx ss
    | Pair (a, b) ->
        chini_list chini_stmt ctx [a; b]
    | Loop {ll_sym = Some sym; ll_body = body; ll_inc = inc} ->
        let loop_lev = (max ctx.cc_loop_level ctx.cc_try_level) + 1 in
        set_property sym (LoopLevel loop_lev);
        let ctx' = chini_stmt {ctx with cc_loop_level = loop_lev} body in
        (* FIXME: This is probably too conservative, however it also
         * catches common cases. We should consider loop body until first
         * continue. But we do not distinguish between break and continue,
         * this might be hard... *)
        let _ = chini_stmt {ctx with cc_loop_level = loop_lev} inc in
        let (abr, br, bbr) = normalize ctx'.cc_this_stmt_initialized in
        let ini =
          if br == -1 or br == loop_lev then 
            bbr
          else 
            (Breaker br) :: bbr in
        {ctx with cc_this_stmt_initialized = ini}
    | Loop {ll_sym = None} -> assert false
    | BreakIfNot e ->
        let ctx = chini_exp ctx e in
        begin
          match e.raw_exp with
          | BoolConst true -> ctx
          | _ ->
            let ini = (Breaker ctx.cc_loop_level) :: 
                      ctx.cc_this_stmt_initialized in
            {ctx with cc_this_stmt_initialized = ini}
        end
    | Break {br_sym = Some sym} ->
        {ctx with cc_this_stmt_initialized = 
                        [Breaker (get_loop_level sym)]}
    | Break {br_sym = None} -> assert false
    | Try t ->
        let try_level = (max ctx.cc_loop_level ctx.cc_try_level) + 1 in
        let inside_ctx = {ctx with cc_try_level = try_level} in
        (* We're not interested in result, as it is always possibility,
         * that control will jump out the try_body. *)
        let _ = chini_stmt inside_ctx t.try_body in
        (* We also do not know if we'll execute any of try_with branches. *)
        let _ = 
          chini_branches ctx (List.map (fun {cc_stmt = s} -> s) t.try_with) in
        (* There is just one thing, we might be sure of: try_finally. *)
        chini_stmt ctx t.try_finally
    | Raise e ->
        let ctx = chini_exp ctx e in
        if ctx.cc_try_level == 1 then
          (* Treat raise() without enclosing try{} same as return. *)
          {ctx with cc_this_stmt_initialized = 
                        Everything :: ctx.cc_this_stmt_initialized}
        else
          (* Otherwise -- as break. *)
          {ctx with cc_this_stmt_initialized = 
                        (Breaker ctx.cc_try_level) 
                                :: ctx.cc_this_stmt_initialized}
    | Switch (e, pl) ->
        let gctx = chini_exp ctx e in
        (* Fortunately we don't have any expressions, involving variables 
         * in patterns (right now...), so focus on stmts only. *)
        chini_branches gctx (List.map (fun {cc_stmt = s} -> s) pl)
    | Let (pl, s) ->
        let ctx = chini_list chini_exp ctx (List.map (fun (p, e) -> e) pl) in
        chini_stmt ctx s
    | If(e, s1, s2) ->
        let ctx = chini_exp ctx e in
        chini_branches ctx [s1; s2]
        
(* Check initialization in one expr. *)
and chini_exp ctx exp =
  match exp.raw_exp with
  (* Fscking special case... otherwise local function cannot be recursive. *)
  | SimpleBinOp (Assign, 
                 {raw_exp = Ref {ref_symbol = Some sym}}, 
                 ({raw_exp = Lambda _} as e2)) ->
      let ctx' = {ctx with cc_initialized =
                                (Symbol sym) :: ctx.cc_initialized} in
      let ini = (chini_exp ctx' e2).cc_this_stmt_initialized in
      {ctx with cc_this_stmt_initialized = (Symbol sym) :: ini}
  | SimpleBinOp (op, e1, e2) ->
      (* We might want to change this, if we define order of evaluation
       * of binary operators arguments... *)
      let ini = (chini_exp ctx e2).cc_this_stmt_initialized in
      let ini =
        match (op, e1.raw_exp) with
          | (Assign, Ref {ref_symbol = Some sym}) -> (Symbol sym) :: ini
          | _ -> (chini_exp ctx e1).cc_this_stmt_initialized @ ini in
      {ctx with cc_this_stmt_initialized = ini}
  | UnionExp {ur_exp = e}
  | DotExp {de_expr = e}
  | SimpleUnOp (_, e) ->
      (* We don't care about the actual operator, union symbol or 
       * whatever... *)
      chini_exp ctx e
  | Ref {ref_symbol = Some sym} ->
      if has_property sym IsUnionField or 
         has_property sym IsInitialized or 
         sym.sym_fnc_level <= 0 then
        ctx
      else
        if List.mem (Symbol sym) ctx.cc_initialized or
           List.mem Everything ctx.cc_initialized then
             ctx
        else
             err (exp.exp_loc, xf "variable %s might be used uninitialized"
                                  sym.sym_name)
  | Ref {ref_symbol = None} -> assert false
  | Call (f, act) ->
      chini_list chini_exp ctx (f :: act)
  | TupleExp tup ->
      chini_list chini_exp ctx tup
  | StructExp sis ->
      chini_list chini_exp ctx (List.map (fun {si_expr = e} -> e) sis)
  | Lambda l ->
      let lctx = chini_stmt ctx l.lambda_body in
      let _ = 
        match l.lambda_ret_type with
        | Void -> ()
        | _ ->
          if List.mem Everything lctx.cc_this_stmt_initialized then
            ()
          else
            err (exp.exp_loc, "this lambda expression can finish without \
                               return") in
      ctx
  | IntConst _
  | BoolConst _
  | FloatConst _
  | StringConst _
  | VoidConst
  | NullConst
  | CrossModuleRef _ -> ctx
;;

(* Check initialization of variables in function. *)
let chini_func ret_type body =
  let ctx = chini_stmt {cc_initialized = []; 
                        cc_this_stmt_initialized = [];
                        cc_loop_level = 1;
                        cc_try_level = 1} body in
  match ret_type with
  | Void -> ()
  | _ ->
    if List.mem Everything ctx.cc_this_stmt_initialized then
      ()
    else
      err (body.stmt_loc, "this function can finish without return")
;;

let typecheck_decl d =
  match d with
      FuncDecl f -> 
        ignore(typecheck_stmt ({ctx_ret_type = f.fn_ret_type; 
                                ctx_tyvars = f.fn_tyvars}, f.fn_body));
        chini_func f.fn_ret_type f.fn_body
    | GlobalDecl _
    | ExternC _
    | FuncProto _
    | TopLevelOpen _
    | DummyDecl
    | SectionDecl _
    | ExceptionDecl _ -> ()
    | TypeDef td -> td_fresh_variables td
;;

let type_of e =
  match e.exp_type with
      Some t -> t
    | None -> assert false
;;
